# coding:utf-8
import typing as t

from makit.web.context import Context

MiddlewareHandler = t.Callable[[Context], t.Awaitable[None]]


class MiddlewareClass(type):
    """"""


class Middleware(metaclass=MiddlewareClass):
    """"""

    async def __call__(self, c: Context):
        await c.next()


class HandlerMiddleware(Middleware):
    def __init__(self, handler):
        self.handler = handler

    async def __call__(self, c: Context):
        await self.handler(c)



# coding:utf-8

import html
import inspect
import traceback

from makit.lib import ospath

FRAME_TEMPLATE = """
<div class="frame">
    <div class="frame-title">
        File {frame_filename}, line <i>{frame_lineno}</i>, in <b>{frame_name}</b>
    </div>
    <div class="source-code">{code_context}</div>
</div>
"""

LINE = """
<div class="source-line {error}">
    <span class="lineno">{lineno}</span> <span class="line-text">{line}</span>
</div>
"""


class ErrorFormatter:
    """

    """

    def format(self, exc) -> str:
        """
        格式化异常
        :param exc:
        :return:
        """
        raise NotImplementedError()


def fmt_text(exc):
    """"""
    return "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))


def fmt_html(exc):
    """"""
    return generate_html(exc)


def generate_html(exc, limit=7):
    traceback_obj = traceback.TracebackException.from_exception(
        exc, capture_locals=True
    )
    exc_html = ""
    exc_traceback = exc.__traceback__
    if exc_traceback is not None:
        frames = inspect.getinnerframes(exc_traceback, limit)
        for frame in reversed(frames):
            exc_html += generate_frame_html(frame)
    error = (
        f"{html.escape(traceback_obj.exc_type.__name__)}: "
        f"{html.escape(str(traceback_obj))}"
    )
    template_file = ospath.join(__file__, '../../static/error-template.html')
    with open(template_file, 'r', encoding='utf-8') as fr:
        template = fr.read()
    content = template.replace('{error}', error).replace('{exc_html}', exc_html)
    # with open('./error.html', 'w', encoding='utf-8') as fw:
    #     fw.write(content)
    return content


def generate_frame_html(frame):
    code_context = "".join(
        format_line(
            index, line, frame.lineno, frame.index
        )
        for index, line in enumerate(frame.code_context or [])
    )

    values = {
        "frame_filename": html.escape(frame.filename),
        "frame_lineno": frame.lineno,
        "frame_name": html.escape(frame.function),
        "code_context": code_context,
    }
    return FRAME_TEMPLATE.format(**values)


def format_line(
        index: int, line: str, frame_lineno: int, frame_index: int
) -> str:
    values = {
        "line": html.escape(line).replace(" ", "&nbsp"),
        "lineno": (frame_lineno - frame_index) + index,
        'error': 'error' if index == frame_index else ''
    }

    return LINE.format(**values)



# coding:utf-8

from makit.web.context import Context
from makit.web.middleware import Middleware


class GZipMiddleware(Middleware):

    async def __call__(self, c: Context):
        # TODO
        pass



# coding:utf-8

from functools import cached_property

from makit.web.context import Context
from makit.web.middleware.base import Middleware

ALL_METHODS = ("DELETE", "GET", "HEAD", "OPTIONS", "PATCH", "POST", "PUT")
SAFE_LISTED_HEADERS = {"Accept", "Accept-Language", "Content-Language", "Content-Type"}


class CORSMiddleware(Middleware):
    def __init__(self, allow_origins=None, allow_credentials=False, allow_methods=None, allow_headers=None):
        super().__init__()
        self._config = dict(
            allow_origins=allow_origins,
            allow_credentials=allow_credentials,
            allow_methods=allow_methods,
            allow_headers=allow_headers
        )

    @cached_property
    def config(self):
        return self.app.config.CORS

    @cached_property
    def allow_all_origins(self):
        return "*" in self.config.allow_origins

    @cached_property
    def allow_all_headers(self):
        return '*' in self.config.allow_headers

    @property
    def preflight_explicit_allow_origin(self):
        return self.allow_all_origins or self.config.allow_credentials

    async def __call__(self, c: Context):
        scope = c.scope
        self.app = c.app
        if scope["type"] != "http":
            return

        method = scope["method"]
        origin = c.headers.get("origin")

        if origin is None:
            return await c.next()

        if method == "OPTIONS" and "access-control-request-method" in c.headers:
            await self.preflight_response(c)
            return

        await c.next()

    def is_allowed_origin(self, origin: str) -> bool:
        if self.allow_all_origins:
            return True
        origin_regex = self.config.allow_origin_regex
        if origin_regex is not None and origin_regex.fullmatch(origin):
            return True
        return origin in self.config.allow_origins

    def is_allowed_method(self, method):
        if '*' in self.config.allow_methods:
            return True
        return method in self.config.allow_methods

    @cached_property
    def preflight_headers(self):
        if '*' in self.config.allow_methods:
            allow_methods = ALL_METHODS
        else:
            allow_methods = self.config.allow_methods
        headers = {
            "Access-Control-Allow-Methods": ", ".join(allow_methods),
            "Access-Control-Max-Age": str(self.config.max_age),
        }
        if "*" in self.config.allow_origins or self.config.allow_credentials:
            headers["Vary"] = "Origin"
        else:
            headers["Access-Control-Allow-Origin"] = "*"
        allow_all_headers = "*" in self.config.allow_headers
        allow_headers = sorted(SAFE_LISTED_HEADERS | set(self.config.allow_headers))
        if allow_headers and not allow_all_headers:
            headers["Access-Control-Allow-Headers"] = ", ".join(allow_headers)
        if self.config.allow_credentials:
            headers["Access-Control-Allow-Credentials"] = "true"

        return headers

    async def preflight_response(self, c: Context):
        headers = c.headers
        requested_origin = headers["origin"]
        requested_method = headers["access-control-request-method"]
        requested_headers = headers.get("access-control-request-headers")

        headers = dict(self.preflight_headers)
        failures = []

        if self.is_allowed_origin(origin=requested_origin):
            if self.preflight_explicit_allow_origin:
                headers["Access-Control-Allow-Origin"] = requested_origin
        else:
            failures.append("origin")

        if not self.is_allowed_method(requested_method):
            failures.append("method")

        if self.allow_all_headers and requested_headers is not None:
            headers["Access-Control-Allow-Headers"] = requested_headers
        elif requested_headers is not None:
            for header in [h.lower() for h in requested_headers.split(",")]:
                if header.strip() not in self.config.allow_headers:
                    failures.append("headers")
                    break

        if failures:
            failure_text = "Disallowed CORS " + ", ".join(failures)
            return await c.respond_text(failure_text, status_code=400, headers=headers)

        return await c.respond_text("OK", status_code=200, headers=headers)

    @staticmethod
    def allow_explicit_origin(headers, origin: str) -> None:
        headers["Access-Control-Allow-Origin"] = origin
        headers.add_vary_header("Origin")



# coding:utf-8

import hashlib
# import json
import os
import sys
import traceback
import typing as t
from email.utils import formatdate
from functools import partial
from mimetypes import guess_type as mimetypes_guess_type
from urllib.parse import quote

import anyio

from makit.lib import json
from makit.web.headers import Headers
from makit.web.types import Scope, Receive, Send
from makit.web.utils.concurrency import iterate_in_threadpool


class Response:
    media_type = None
    charset = 'utf-8'

    def __init__(self, content=None, status_code=200, headers=None, media_type=None):
        self.status_code = status_code
        self.media_type = media_type or self.__class__.media_type
        self.body = self.render(content)
        self.headers = headers or Headers(charset=self.charset)

    def render(self, content: t.Any) -> bytes:
        if content is None:
            return b""
        if isinstance(content, bytes):
            return content
        if hasattr(content, 'encode'):
            return content.encode(self.charset)
        return content.encode(self.charset)

    def build_headers(self, scope=None):
        headers = Headers(scope=scope)
        headers.charset = self.charset
        headers.update(**self.headers)
        if not (self.status_code < 200 or self.status_code in (204, 304)):
            content_length = str(len(self.body))
            headers['content-length'] = content_length
        content_type = self.media_type
        if content_type is not None:
            if content_type.startswith("text/"):
                content_type += "; charset=" + self.charset
            headers['content-type'] = content_type
        origin = headers.get('Origin')
        headers['Access-Control-Allow-Origin'] = origin
        return headers

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        try:
            headers = self.build_headers(scope)
            await send(
                {
                    "type": "http.response.start",
                    "status": self.status_code,
                    "headers": headers.raw_data,
                }
            )
            await send({"type": "http.response.body", "body": self.body})
        except:
            print(traceback.format_exc())


class PlainTextResponse(Response):
    media_type = "text/plain"


class HtmlResponse(Response):
    media_type = "text/html"


class JsonResponse(Response):
    media_type = "application/json"

    def __init__(
            self,
            content: t.Any,
            status_code: int = 200,
            headers: t.Optional[t.Dict[str, str]] = None
    ):
        super().__init__(content, status_code, headers, None)

    def render(self, content: t.Any) -> bytes:
        serialized = json.dumps(content).encode('utf-8')
        return serialized


class FileResponse(Response):
    chunk_size = 64 * 1024

    def __init__(
            self,
            path: t.Union[str, "os.PathLike[str]"],
            status_code: int = 200,
            headers: t.Optional[t.Mapping[str, str]] = None,
            media_type: t.Optional[str] = None,
            filename: t.Optional[str] = None,
            stat_result: t.Optional[os.stat_result] = None,
            method: t.Optional[str] = None,
            content_disposition_type: str = "attachment"
    ):
        super().__init__(None, status_code, headers, media_type)
        self.path = path
        self.status_code = status_code
        self.filename = filename
        self.send_header_only = method is not None and method.upper() == "HEAD"
        if media_type is None:
            media_type = guess_type(filename or path)[0]
            media_type = media_type or "application/octet-stream"
        self.media_type = media_type
        if self.filename is not None:
            content_disposition_filename = quote(self.filename)
            if content_disposition_filename != self.filename:
                content_disposition = "{}; filename*=utf-8''{}".format(
                    content_disposition_type, content_disposition_filename
                )
            else:
                content_disposition = '{};filename="{}"'.format(
                    content_disposition_type, self.filename
                )
            self.headers["content-disposition"] = content_disposition
        self.stat_result = stat_result
        if stat_result is not None:
            self.set_stat_headers(stat_result)

    def set_stat_headers(self, stat_result: os.stat_result) -> None:
        content_length = str(stat_result.st_size)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        etag = md5_hexdigest(etag_base.encode(), usedforsecurity=False)

        self.headers["content-length"] = content_length
        self.headers["last-modified"] = last_modified
        self.headers["etag"] = etag

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        origin = Headers(scope=scope)['origin']
        headers = self.build_headers()
        headers['Access-Control-Allow-Origin'] = origin
        headers.remove('token')
        if self.stat_result is None:
            try:
                stat_result = await anyio.to_thread.run_sync(os.stat, self.path)
                content_length = str(stat_result.st_size)
                last_modified = formatdate(stat_result.st_mtime, usegmt=True)
                etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
                etag = md5_hexdigest(etag_base.encode(), usedforsecurity=False)

                headers["content-length"] = content_length
                headers["last-modified"] = last_modified
                headers["etag"] = etag
            except FileNotFoundError:
                raise RuntimeError(f"File at path {self.path} does not exist.")
            else:
                mode = stat_result.st_mode
                if not os.stat.S_ISREG(mode):
                    raise RuntimeError(f"File at path {self.path} is not a file.")
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": headers.raw_data,
            }
        )
        if self.send_header_only:
            await send({"type": "http.response.body", "body": b"", "more_body": False})
        else:
            async with await anyio.open_file(self.path, mode="rb") as file:
                more_body = True
                size = 0
                while more_body:
                    chunk = await file.read(self.chunk_size)
                    size += len(chunk)
                    more_body = len(chunk) == self.chunk_size
                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": more_body,
                        }
                    )


class RedirectResponse(Response):
    def __init__(
            self,
            url: str,
            status_code: int = 307,
            headers: t.Optional[t.Mapping[str, str]] = None
    ):
        super().__init__(
            content=b"", status_code=status_code, headers=headers
        )
        self.headers["location"] = quote(str(url), safe=":/%#?=@[]!$&'()*+,;")


Content = t.Union[str, bytes]
SyncContentStream = t.Iterator[Content]
AsyncContentStream = t.AsyncIterable[Content]
ContentStream = t.Union[AsyncContentStream, SyncContentStream]


class StreamingResponse(Response):
    body_iterator: AsyncContentStream

    def __init__(
            self, content: ContentStream,
            status_code: int = 200,
            headers: t.Optional[t.Mapping[str, str]] = None,
            media_type: t.Optional[str] = None
    ):
        super().__init__(None, status_code, headers, media_type)
        if isinstance(content, t.AsyncIterable):
            self.body_iterator = content
        else:
            self.body_iterator = iterate_in_threadpool(content)
        self.status_code = status_code
        self.media_type = self.media_type if media_type is None else media_type

    async def listen_for_disconnect(self, receive: Receive) -> None:
        while True:
            message = await receive()
            if message["type"] == "http.disconnect":
                break

    async def stream_response(self, scope: Scope, send: Send) -> None:
        headers = self.build_headers(scope)
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": headers.raw_data,
            }
        )
        async for chunk in self.body_iterator:
            if not isinstance(chunk, bytes):
                chunk = chunk.encode(self.charset)
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async with anyio.create_task_group() as task_group:
            async def wrap(func: "t.Callable[[], t.Awaitable[None]]") -> None:
                await func()
                await task_group.cancel_scope.cancel()

            task_group.start_soon(wrap, partial(self.stream_response, scope, send))
            await wrap(partial(self.listen_for_disconnect, receive))


def guess_type(
        url: t.Union[str, "os.PathLike[str]"], strict: bool = True
) -> t.Tuple[t.Optional[str], t.Optional[str]]:
    if sys.version_info < (3, 8):  # pragma: no cover
        url = os.fspath(url)
    return mimetypes_guess_type(url, strict)


def md5_hexdigest(data: bytes, *, usedforsecurity: bool = True) -> str:
    return hashlib.md5(data).hexdigest()



# coding:utf-8
import traceback

from makit.web.context import Context
from makit.web.exceptions import HttpUnauthorizedException, HttpNotFoundException, HttpForbiddenException
from makit.web.middleware.base import Middleware
from makit.web.utils import error


class HandleServerError(Middleware):
    def __init__(self, error_handler=None):
        super().__init__()
        self.error_handler = error_handler

    async def __call__(self, c: Context):
        app = c.app
        try:
            await c.next()
            if not c.response_completed:
                await c.respond_json(dict(message='OK', succeeded=True))
        except HttpUnauthorizedException as e:
            await c.respond_json(str(e), status_code=401)
        except HttpNotFoundException as e:
            await c.respond_json(str(e), status_code=404)
        except HttpForbiddenException as e:
            await c.respond_json(str(e), status_code=e.status_code)
        except Exception as exc:
            try:
                c.exception = exc
                logger = app.logger
                tb = traceback.format_exc()
                logger.debug(tb)
                sys_config = app.config.APP
                await app.handle_server_error(c, exc)
                debug = sys_config.debug
                if debug:
                    return await self.debug_response(c, exc)
                else:
                    await c.respond_json('Internal Server Error', status_code=500)
            except:
                await c.respond_json('Internal Server Error', status_code=500)

    async def debug_response(self, c: Context, exc):
        accept = c.headers.get("accept", "")
        if "text/html" in accept:
            content = error.generate_html(exc)
            return await c.respond_html(content, status_code=500)
        content = error.fmt_text(exc)
        return await c.respond_text(content, status_code=500)



# coding:utf-8

import asyncio
import os
import traceback
from datetime import datetime, UTC, timedelta

import jwt
import uvicorn
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from logzero import setup_logger

from makit.lib import Config
from makit.lib.command import AppCommand
from makit.web import default_config, status
from makit.web.context import Context
from makit.web.convert import CONVERTERS, PathConverter
from makit.web.exceptions import HttpException
from makit.web.middleware.error import HandleServerError
from makit.web.routing import Router
from makit.web.static import index, static_file
from makit.web.types import Scope, Receive, Send


class Application(Router):
    __context__ = Context
    __scheduler_class__ = AsyncIOScheduler

    def __init__(self, prefix='', error_handler=None, config=None, logger=None, prog=None):
        super().__init__(prefix)
        self._base_dir = os.getcwd()
        self.config = Config(default_config).use(config)
        self.logger = logger or setup_logger('makit-web')
        self.lifespan = DefaultLifespan(self)
        self._path_converters = CONVERTERS
        self.scheduler = self.__scheduler_class__() if self.__scheduler_class__ else None
        self.use(HandleServerError, error_handler=error_handler)
        self._command = AppCommand(prog)

    @property
    def base_dir(self):
        return self._base_dir

    @property
    def static_dirs(self):
        dirs = self.config.APP.static_dirs
        if not dirs:
            dirs = [os.path.join(self.base_dir, 'static')]
        return dirs

    def get_config(self, key):
        return self.config.get(key)

    def jwt_encode(self, payload: dict) -> str:
        secret_key = self.config.JWT.get('secret_key')
        if not secret_key:
            raise HttpException(detail='missing jwt config: secret_key')
        algorithm = self.config.JWT.get('algorithm', 'HS256')
        payload['exp'] = datetime.now(UTC) + self.config.JWT.get('expiration', timedelta(days=7))
        audience = self.config.JWT.get('audience', None)
        if audience:
            payload['audience'] = audience
        issuer = self.config.JWT.get('issuer', None)
        if issuer:
            payload['issuer'] = issuer
        return jwt.encode(payload, secret_key, algorithm=algorithm)

    def jwt_decode(self, token: str) -> dict:
        secret_key = self.config.JWT.get('secret_key')
        algorithm = self.config.JWT.get('algorithm', 'HS256')
        payload = jwt.decode(token, secret_key, algorithms=algorithm, verify=True)
        return payload

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        scope_type = scope.get('type')
        if scope_type == 'lifespan':
            await self._lifespan(scope, receive, send)
        elif scope_type == 'http':
            c = self.__context__(scope, receive, send)
            c.app = self
            route = self.matches(c)
            if not route:
                path = scope.get('path')
                await c.respond_json(f'route not found: {path}', status_code=status.HTTP_404_NOT_FOUND)
                return
            await c.next()

    def run(self, **kwargs):
        sys_config = self.config.APP
        host = kwargs.pop('host', sys_config.get('host', '127.0.0.1'))
        port = kwargs.pop('port', sys_config.get('port', 8000))

        self.add_get('/', index)
        # self.router.add_get('/media/{path:path}', media_resource, methods=['GET'])
        # self.router.add_get('/static/{path:path}', static_file)
        self.add_get('/assets/{path:path}', static_file)
        if self.config.APP.debug:
            for route in self.routes:
                self.logger.info(f'route: {route.path_regex.pattern}, {route.methods}, {route.name}')

        uvicorn.run(self, ws_max_size=20 * 1024 * 1024, host=host, port=port, **kwargs)

    def startup(self):
        return self.lifespan.startup()

    def shutdown(self):
        return self.lifespan.shutdown()

    def add_path_converter(self, token: str, cvt: PathConverter):
        self._path_converters[token] = cvt

    def enable_open_api(self):
        """ 启用 OpenAPI """

    def command(self, default=False):
        """
        命令装饰器

        示例：

        app = Application()

        @pp.command(default=True)
        def serve():
            app.run()

        app.run_from_command()

        命令行：python main.py serve  或  python main.py

        :param default: 是否默认命令
        :return:
        """
        return self._command.command(default)

    def run_from_command(self):
        """
        从命令行启动
        :return:
        """
        self._command.run()

    async def _lifespan(self, scope: Scope, receive: Receive, send: Send) -> None:
        started = False
        await receive()
        try:
            async with self.lifespan():
                await send({"type": "lifespan.startup.complete"})
                started = True
                await receive()
        except BaseException:
            exc_text = traceback.format_exc()
            if started:
                await send({"type": "lifespan.shutdown.failed", "message": exc_text})
            else:
                await send({"type": "lifespan.startup.failed", "message": exc_text})
            raise
        else:
            await send({"type": "lifespan.shutdown.complete"})


class DefaultLifespan:
    def __init__(self, app):
        self.app = app
        self._startup_handlers = []
        self._shutdown_handlers = []

    async def __aenter__(self) -> None:
        if self.app.scheduler and not self.app.scheduler.running:
            self.app.scheduler.start()
        for handler in self._startup_handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler()
            else:
                handler()

    async def __aexit__(self, *exc_info: object) -> None:
        for handler in self._shutdown_handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler()
            else:
                handler()

    def __call__(self):
        return self

    def startup(self):
        def deco(func):
            self._startup_handlers.append(func)
            return func

        return deco

    def shutdown(self):
        def deco(func):
            self._shutdown_handlers.append(func)
            return func

        return deco



# coding:utf-8

import re
import typing as t
from functools import cached_property
from urllib.parse import parse_qsl

from python_multipart.multipart import parse_options_header

from makit.lib import json
from makit.web import status
from makit.web.exceptions import DisconnectError
from makit.web.file import UploadFile
from makit.web.headers import Headers
from makit.web.parser import MultiPartParser, FormParser
from makit.web.response import JsonResponse, HtmlResponse, Response, FileResponse, PlainTextResponse
from makit.web.types import Scope, Receive, Send
from makit.web.ws import WebSocket


class QueryParams(dict):
    def __init__(self, *args):
        assert len(args) < 2, "Too many arguments."
        value = args[0] if args else []
        if isinstance(value, bytes):
            value = value.decode("latin-1")
        value = parse_qsl(value, keep_blank_values=True)
        d = {item[0]: self._convert(item[1]) for item in value}
        super().__init__(**d)

    @classmethod
    def _convert(cls, value):
        if re.match(r'^\d+$', value):
            return int(value)
        elif re.match(r'\d+(\.\d+)?$', value):
            return float(value)
        elif value in ['null', 'undefined']:
            return None
        elif value.lower() == 'false':
            return False
        elif value.lower() == 'true':
            return True
        return value


class Context:

    def __init__(self, scope: Scope, receive: Receive, send: Send):
        self.app = None
        self.scope = scope
        self._receive = receive
        self._send = send
        self._body, self._json, self._form = None, None, None
        self._headers = Headers(scope=scope)
        self.response_completed = False
        self._middlewares = None  # 中间件
        self._next_middleware = None  # 下一个执行的中间件

    @cached_property
    def query(self):
        return QueryParams(self.scope["query_string"])

    @cached_property
    def params(self):
        path_params = self.scope['route'].path_params
        return path_params

    @property
    def headers(self):
        return self._headers

    @cached_property
    def method(self):
        """request method"""
        return self.scope['method']

    @property
    def route(self):
        return self.scope.get('route')

    async def body(self) -> bytes:
        if self._body is None:
            chunks = []
            async for chunk in self._stream():
                chunks.append(chunk)
            self._body = b"".join(chunks)
        return self._body

    async def json(self):
        if self._json is None:
            content_type_header = self.headers.get("Content-Type")
            if content_type_header:
                content_type, options = parse_options_header(content_type_header)
                if content_type != b'application/json':
                    return {}
            data = await self.body()
            data = data.decode('utf-8') or '{}'
            data = json.loads(data)
            self._json = data
        return self._json

    async def form(self):
        if self._form is None:
            content_type_header = self.headers.get("Content-Type")
            content_type, options = parse_options_header(content_type_header)
            if content_type == b"multipart/form-data":
                multipart_parser = MultiPartParser(self.headers, self._stream())
                self._form = await multipart_parser.parse()
            elif content_type == b"application/x-www-form-urlencoded":
                form_parser = FormParser(self.headers, self._stream())
                self._form = await form_parser.parse()
            else:
                self._form = []
        return self._form

    async def files(self):
        form_items = await self.form()
        return [item for name, item in form_items if isinstance(item, UploadFile)]

    def handle_api_output(self, output):
        """ 用于在子类中实现对API输出的统一处理 """
        return output

    async def respond(self, content, status_code=status.HTTP_200_OK, headers=None):
        if self.response_completed:
            return
        response = Response(content, status_code=status_code, headers=headers)
        await response(self.scope, self._receive, self._send)
        self.response_completed = True

    async def respond_text(self, content, status_code=status.HTTP_200_OK, headers=None):
        if self.response_completed:
            return
        response = PlainTextResponse(content, status_code=status_code, headers=headers)
        await response(self.scope, self._receive, self._send)
        self.response_completed = True

    async def respond_json(self, data, status_code=200, headers=None):
        if self.response_completed:
            return
        data = json.predumps(data)
        response = JsonResponse(data, status_code=status_code, headers=headers)
        await response(self.scope, self._receive, self._send)
        self.response_completed = True

    async def respond_html(self, content, status_code=status.HTTP_200_OK, headers=None):
        if self.response_completed:
            return
        response = HtmlResponse(content, status_code=status_code, headers=headers)
        await response(self.scope, self._receive, self._send)
        self.response_completed = True

    async def respond_file(
            self,
            file,
            filename=None,
            headers=None,
            content_disposition_type='attachment',
            stat_result=None,
            status_code=200
    ):
        if self.response_completed:
            return
        response = FileResponse(
            file,
            filename=filename,
            status_code=status_code,
            headers=headers,
            stat_result=stat_result,
            content_disposition_type=content_disposition_type
        )
        await response(self.scope, self._receive, self._send)
        self.response_completed = True

    async def unauthorized(self, message, headers=None):
        await self.respond_json(message, status_code=status.HTTP_401_UNAUTHORIZED, headers=headers)

    def websocket(self):
        ws = WebSocket(self.scope, self._receive, self._send)
        return ws

    async def next(self):
        """
        执行下一个中间件，只能在中间件中被调用
        """
        if self._middlewares is None:
            self._middlewares = []
            for middleware_class, args, kwargs in self.route.middlewares:
                middleware = middleware_class(*args, **kwargs)
                self._middlewares.insert(0, middleware)

        if len(self._middlewares) > 0:
            middleware = self._middlewares.pop()
            await middleware(self)
        else:
            await self.route(self)

    async def _stream(self) -> t.AsyncGenerator[bytes, None]:
        if self._body is not None:
            yield self._body
            yield b""
            return
        while True:
            message = await self._receive()
            if message["type"] == "http.request":
                body = message.get("body", b"")
                if body:
                    yield body
                if not message.get("more_body", False):
                    break
            elif message["type"] == "http.disconnect":
                raise DisconnectError('http disconnect')
        yield b""



# coding:utf-8

import os
import tempfile
import typing as t

from makit.web.headers import Headers
from makit.web.utils.concurrency import run_in_threadpool


class UploadFile:
    """
    An uploaded file included as part of the request data.
    """

    spool_max_size = 1024 * 1024
    file: t.BinaryIO
    headers: Headers

    def __init__(
            self,
            filename: str,
            file: t.Optional[t.BinaryIO] = None,
            content_type: str = "",
            *,
            headers: t.Optional[Headers] = None,
    ) -> None:
        self.filename = filename
        self.content_type = content_type
        if file is None:
            self.file = tempfile.SpooledTemporaryFile(max_size=self.spool_max_size)  # type: ignore  # noqa: E501
        else:
            self.file = file
        self.headers = headers or Headers()

    @property
    def _in_memory(self) -> bool:
        rolled_to_disk = getattr(self.file, "_rolled", True)
        return not rolled_to_disk

    async def write(self, data: bytes) -> None:
        if self._in_memory:
            self.file.write(data)
        else:
            await run_in_threadpool(self.file.write, data)

    async def read(self, size: int = -1) -> bytes:
        if self._in_memory:
            return self.file.read(size)
        return await run_in_threadpool(self.file.read, size)

    async def seek(self, offset: int) -> None:
        if self._in_memory:
            self.file.seek(offset)
        else:
            await run_in_threadpool(self.file.seek, offset)

    async def close(self) -> None:
        if self._in_memory:
            self.file.close()
        else:
            await run_in_threadpool(self.file.close)

    async def save(self, filename=None):
        if not filename:
            filename = self.filename
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, 'wb') as f:
            f.write(await self.read())


class File:
    pass



# coding:utf-8

import os

from makit.lib import ospath
from makit.web.context import Context


async def index(c: Context):
    app = c.app
    static_dir = os.path.join(app.config.SYSTEM.base_dir, 'static')
    html = os.path.join(static_dir, 'index.html')
    with open(html, 'r', encoding='utf-8') as h:
        content = h.read()
    return await c.respond_html(content)


async def static_file(c: Context):
    app = c.app
    path = c.scope['path']

    if path.startswith('/static'):
        path = path[7:]
    base_dir = app.base_dir
    filepath = ospath.join(base_dir, 'static', path)
    if not os.path.exists(filepath):
        filepath = ospath.join(base_dir, path)
        if not os.path.exists(filepath):
            return await c.respond_json('Not found', status_code=404)
    if path.endswith('.js'):
        with open(filepath, 'r', encoding='utf-8') as f:
            content = f.read()
        return await c.respond(content=content, headers={'content-type': 'text/javascript'})
    else:
        with open(filepath, 'rb') as f:
            content = f.read()
    if 'temp' in filepath:
        os.remove(filepath)
    return await c.respond(content=content)



# coding:utf-8


class PathConverter:
    regex = None

    def convert(self, value: str):
        return value


class StrConverter(PathConverter):
    regex = '[^/]+'


class IntConverter(PathConverter):
    regex = r'\d+'

    def convert(self, value) -> int:
        return int(value)


class BoolConverter(PathConverter):
    regex = '(1|true|0|false)'

    def convert(self, value: str) -> bool:
        value = value.lower()
        if value in ['1', 'true']:
            return True
        return False


class FloatConvertor(PathConverter):
    regex = r"[0-9]+(\.[0-9]+)?"

    def convert(self, value: str) -> float:
        return float(value)


class PathConvertor(PathConverter):
    regex = ".*"


CONVERTERS = {
    'str': StrConverter,
    'int': IntConverter,
    'float': FloatConvertor,
    'path': PathConvertor,
    'bool': BoolConverter
}



# coding:utf-8

import typing as t

from makit.lib.encoding import decode, encode


class Headers(t.Mapping[str, str]):
    """
    Http Headers
    """

    def __init__(self, raw=None, scope=None, charset='"latin-1"', **kwargs):
        self._raw_data = raw or []
        self.charset = charset
        if scope:
            for k, v in scope.get('headers'):
                self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    @property
    def raw_data(self):
        return self._raw_data

    def keys(self):
        return [decode(k, encoding=self.charset) for k, v in self._raw_data]

    def values(self):
        return [decode(v, encoding=self.charset) for k, v in self._raw_data]

    def items(self):
        return [
            (decode(key, self.charset), decode(value, self.charset))
            for key, value in self._raw_data
        ]

    def get(self, key: t.Union[str, bytes], default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, key: str, value):
        value = decode(value, encoding=self.charset)
        if key not in self.keys():
            self[key] = value
        return value

    def add_vary_header(self, vary: str) -> None:
        existing = self.get("vary")
        if existing is not None:
            vary = ", ".join([existing, vary])
        self["vary"] = vary

    def update(self, **kwargs):
        for k, v in kwargs.items():
            self[k] = v

    def remove(self, key):
        for item in self._raw_data:
            k = decode(item[0], encoding=self.charset)
            if k == key:
                self._raw_data.remove(item)
                break

    def __contains__(self, item: t.Union[str, bytes]):
        item = encode(item.lower(), encoding=self.charset)
        for k, v in self._raw_data:
            if k == item:
                return True
        return False

    def __setitem__(self, key: t.Union[str, bytes], value):
        key = key.lower()
        key = encode(key, encoding=self.charset)
        for item in self._raw_data:
            if item[0] == key:
                self._raw_data.remove(item)
                break
        self._raw_data.append(
            (key, encode(value, encoding=self.charset))
        )

    def __getitem__(self, k: t.Union[str, bytes]) -> str:
        item = encode(k.lower(), encoding=self.charset)
        if self._raw_data:
            for k, v in self._raw_data:
                if k == item:
                    return decode(v, encoding=self.charset)
        raise KeyError(k)

    def __len__(self) -> int:
        return len(self._raw_data)

    def __iter__(self):
        for k, v in self._raw_data:
            key = decode(k, encoding=self.charset)
            value = decode(v, encoding=self.charset)
            yield key, value



# coding:utf-8
import fnmatch
import typing as t
from functools import cached_property

from makit.web import status
from makit.web.context import Context
from makit.web.middleware import Middleware


class HostMiddleware(Middleware):

    def __init__(
            self,
            allowed_hosts: t.List[str] | None = None,
            disallowed_hosts: t.List[str] | None = None,
    ):
        self.allowed_hosts = allowed_hosts or ['*']
        self.disallowed_hosts = disallowed_hosts or []

    @cached_property
    def allow_all(self):
        return '*' in self.allowed_hosts

    async def __call__(self, c: Context):
        host = c.headers.get("host").split(":")[0]
        allowed = True
        for pattern in self.disallowed_hosts:
            if fnmatch.fnmatch(host, pattern):
                allowed = False
                break

        if not allowed:
            await c.respond_text('disallowed host!', status_code=status.HTTP_403_FORBIDDEN)
            return

        allowed = self.allow_all
        for pattern in self.allowed_hosts:
            if fnmatch.fnmatch(host, pattern):
                allowed = True
                break

        if allowed:
            await c.next()
        else:
            await c.respond_text('disallowed host!', status_code=status.HTTP_403_FORBIDDEN)



# coding:utf-8

import asyncio
import functools
import typing as t

import anyio

T = t.TypeVar("T")


async def run_in_threadpool(func, *args, **kwargs):
    if kwargs:
        func = functools.partial(func, **kwargs)
    return await anyio.to_thread.run_sync(func, *args)


class _StopIteration(Exception):
    pass


def _next(iterator: t.Iterator[T]) -> T:
    # We can't raise `StopIteration` from within the threadpool iterator
    # and catch it outside that context, so we coerce them into a different
    # exception type.
    try:
        return next(iterator)
    except StopIteration:
        raise StopIteration()


async def iterate_in_threadpool(
        iterator: t.Iterator[T],
) -> t.AsyncIterator[T]:
    while True:
        try:
            yield await anyio.to_thread.run_sync(_next, iterator)
        except StopIteration:
            break


def is_async_callable(obj: t.Any) -> bool:
    while isinstance(obj, functools.partial):
        obj = obj.func

    return asyncio.iscoroutinefunction(obj) or (
            callable(obj) and asyncio.iscoroutinefunction(obj.__call__)
    )



# coding:utf-8

import asyncio
import inspect
import re
import typing as t
from functools import cached_property

import pydantic

from makit.lib import inspect
from makit.web import status
from makit.web.context import Context
from makit.web.convert import CONVERTERS
from makit.web.exceptions import InvalidMiddlewareError

__all__ = [
    'Route',
    'Router'
]

from makit.web.middleware.base import HandlerMiddleware, Middleware, MiddlewareHandler, MiddlewareClass

PARAM_REGEX = re.compile("{([a-zA-Z_][a-zA-Z0-9_]*)(?::([a-zA-Z_][a-zA-Z0-9_]*))?}")


class Router:

    def __init__(self, prefix='', parent=None):
        self.prefix = prefix or ''
        self._routes: t.List['Route'] = []
        self.parent: t.Optional['Router'] = parent
        self.children: t.List['Router'] = []
        self.middlewares = []

    @cached_property
    def path(self):
        result = self.prefix
        r = self
        while r.parent:
            result = r.parent.prefix + result
            r = r.parent
        return result

    @property
    def routes(self):
        return list(self.__iter__())

    def matches(self, c: Context) -> t.Optional['Route']:
        path = c.scope.get('path')
        match = re.compile('^' + self.path).match(path)
        if not match:
            return None
        # c.use_middleware(*self.middlewares)
        for route in self._routes:
            match = route.matches(c)
            if match:
                return c.route

        for child in self.children:
            route = child.matches(c)
            if route:
                return route
        return None

    def sub_router(self, prefix=''):
        router = Router(prefix=prefix, parent=self)
        self.children.append(router)
        return router

    def include(self, router):
        router.parent = self
        self.children.append(router)

    def use(self, middleware: MiddlewareHandler | MiddlewareClass, *args, **kwargs):
        if inspect.isroutine(middleware):
            self.middlewares.append((HandlerMiddleware, [middleware], dict()))
        elif inspect.isclass(middleware) and issubclass(middleware, Middleware):
            self.middlewares.append((middleware, args, kwargs))
        else:
            raise InvalidMiddlewareError('Middleware must be MiddlewareHandler or Middleware class')

    def route(self, path: str, methods: t.List[str], **kwargs):

        def deco(f):
            self.add_route(path, f, methods=methods, **kwargs)

        return deco

    def get(self, path: str, **kwargs):
        return self.route(path, ['GET'], **kwargs)

    def post(self, path: str, **kwargs):
        return self.route(path, ['POST'], **kwargs)

    def put(self, path: str, **kwargs):
        return self.route(path, ['PUT'], **kwargs)

    def patch(self, path: str, **kwargs):
        return self.route(path, ['PATCH'], **kwargs)

    def delete(self, path: str, **kwargs):
        return self.route(path, ['DELETE'], **kwargs)

    def add_route(self, path: str, endpoint, methods: t.List[str], **kwargs):
        route = Route(path, endpoint, methods, router=self, **kwargs)
        self._routes.append(route)
        return route

    def add_get(self, path: str, endpoint, **kwargs):
        return self.add_route(path, endpoint, methods=['GET'], **kwargs)

    def add_post(self, path: str, endpoint, **kwargs):
        return self.add_route(path, endpoint, methods=['POST'], **kwargs)

    def add_put(self, path: str, endpoint, **kwargs):
        return self.add_route(path, endpoint, methods=['PUT'], **kwargs)

    def add_patch(self, path: str, endpoint, **kwargs):
        return self.add_route(path, endpoint, methods=['PATCH'], **kwargs)

    def add_delete(self, path: str, endpoint, **kwargs):
        return self.add_route(path, endpoint, methods=['DELETE'], **kwargs)

    def add_head(self, path: str, endpoint, **kwargs):
        return self.add_route(path, endpoint, methods=['HEAD'], **kwargs)

    def __repr__(self):
        return f'<Router {self.path}>'

    def __iter__(self):
        for route in self._routes:
            yield route
        for router in self.children:
            for route in router:
                yield route


class Route:
    def __init__(self, path: str, handler, methods, router=None, **kwargs):
        self.path = path
        self.handler = handler
        self.methods = methods
        if 'GET' in methods:
            self.methods.append('HEAD')
        self._data = kwargs
        self.param_converters = {}
        self.path_params = dict()
        self.router = router
        self.__handler_func = inspect.Method(handler)
        # self.openapi_params = []
        # for arg in f.args:
        #     if issubclass(arg.type, pydantic.BaseModel):
        #         self.openapi_params.append(arg)

    @cached_property
    def name(self):
        name = self._data.get('name')
        if name:
            return name
        return inspect.obj_name(self.handler)

    @cached_property
    def path_regex(self):
        match = PARAM_REGEX.search(self.path)
        i = 0
        regex = ''
        if not match:
            return re.compile(self.router.path + self.path + '$')
        for match in PARAM_REGEX.finditer(self.path):
            p, t = match.groups("str")  # TODO
            if t not in CONVERTERS:
                raise RouteError(f'Unsupported path converter type: {t}')
            if p in self.param_converters:
                raise RouteError(f'Duplicate path param: {p}')
            self.param_converters[p] = cvt = CONVERTERS[t]
            regex += self.path[i: match.start()]
            regex += f"(?P<{p}>{cvt.regex})"
            i = match.end()
        regex += self.path[i:] + '$'
        regex = (self.router.path + regex).replace('//', '/')
        return re.compile(regex)

    @property
    def data(self):
        """
        Data bound to route
        :return:
        """
        return self._data

    @cached_property
    def middlewares(self):
        router = self.router
        stack = list(router.middlewares)
        while router.parent:
            for middleware in reversed(router.parent.middlewares):
                stack.insert(0, middleware)
            router = router.parent
        return stack

    def matches(self, c: Context):
        path = c.scope['path']
        match = self.path_regex.fullmatch(path)
        if match and (c.method in self.methods or c.method == 'OPTIONS'):
            path_params = match.groupdict()
            for key, value in path_params.items():
                path_params[key] = self.param_converters[key]().convert(value)
            self.path_params = path_params
            c.scope['route'] = self
            return True
        return False

    async def __call__(self, c: Context):
        scope = c.scope
        method = scope['method']
        if method != "OPTIONS" and scope["method"] not in self.methods:
            headers = {"Allow": ", ".join(self.methods)}
            await c.respond_text(
                "Method Not Allowed",
                status_code=status.HTTP_405_METHOD_NOT_ALLOWED,
                headers=headers
            )
        else:
            if not self.__handler_func:
                self.__handler_func = inspect.Method(self.handler)
            f = self.__handler_func
            args, kwargs = [], {}
            json_data = await c.json()
            kwargs.update(json_data)
            kwargs.update(c.query)
            kwargs.update(c.route.path_params)
            for p in f.args:
                if p.name == 'self':
                    args.append(object.__new__(f.cls))
                if not inspect.isclass(p.type):
                    continue
                if issubclass(p.type, pydantic.BaseModel):
                    kwargs[p.name] = p.type(**kwargs)
                elif issubclass(p.type, Context):
                    kwargs[p.name] = c
            if asyncio.iscoroutinefunction(self.handler):
                output = await inspect.async_run(self.handler, *args, **kwargs)
            else:
                output = inspect.run(self.handler, *args, **kwargs)
            output = c.handle_api_output(output)
            if not c.response_completed:
                if output is not None:
                    await c.respond_json(output)
                else:
                    await c.respond_json(dict(message='OK'))

    def __getattr__(self, item):
        value = self._data.get(item, None)
        return value

    def __getitem__(self, item):
        return self._data.get(item, None)

    def __repr__(self):
        return f'<Route {self.path_regex}>'


class RouteError(Exception):
    """"""


class RouteDataError(Exception):
    """"""



# coding:utf-8

from makit.web.context import Context
from makit.web.middleware.base import Middleware


class JWTMiddleware(Middleware):

    async def __call__(self, c: Context):
        token = c.headers.get('token')
        if not token:
            token = c.headers.get('authorization')
        if not token:
            await c.unauthorized('http headers should has TOKEN or AUTHORIZATION')
            return
        try:
            payload = c.app.jwt_decode(token)
            c.scope['payload'] = payload
        except Exception:
            await c.unauthorized('invalid token')

        await c.next()



from setuptools import setup, find_packages

packages = find_packages()

setup(
    name='makitdone-web',
    version='0.0.1',
    packages=[
        'makit.web',
        'makit.web.middleware',
        'makit.web.utils',
        'makit.web.samples'
    ],
    namespace_package=['makit', 'makit.web'],
    install_requires=[
        'uvicorn~=0.35.0',
        'logzero~=1.7.0',
        'PyJWT~=2.10.1',
        'pydantic~=2.11.7',
        'anyio~=4.10.0',
        'makitdone-lib~=2.0.1',
        'python-multipart~=0.0.20',
        'APScheduler~=3.11.0'
    ],
    python_requires='>=3.11',
    url='https://gitee.com/makitdone/makitdone-web',
    license='MIT',
    author='liangchao',
    author_email='liang20201101@163.com',
    description='基于uvicorn的web框架'
)



# coding:utf-8

import http
import typing as t

from makit.web import status


class ConfigException(Exception):
    """"""


class InvaidConfigException(ConfigException):
    """"""


class DisconnectError(Exception):
    """"""


class HttpException(Exception):
    def __init__(
            self,
            status_code: t.Optional[int] = None,
            detail: t.Optional[str] = None,
            headers: t.Optional[dict] = None,
    ) -> None:
        status_code = status_code or status.HTTP_500_INTERNAL_SERVER_ERROR
        if detail is None:
            detail = http.HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.detail = detail
        self.headers = headers

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"


class HttpNotFoundException(HttpException):
    def __init__(self, detail=None, headers=None):
        super().__init__(status_code=status.HTTP_404_NOT_FOUND, detail=detail, headers=headers)


class HttpForbiddenException(HttpException):
    def __init__(self, detail=None, headers=None):
        super().__init__(status_code=status.HTTP_403_FORBIDDEN, detail=detail, headers=headers)


class HttpUnauthorizedException(HttpException):
    def __init__(self, detail=None, headers=None):
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED, detail=detail, headers=headers)


class InvalidMiddlewareError(Exception):
    """"""
